/*
 * Please do not edit this file.
 * It was generated using rpcgen.
 */

#include "config.h"
#include "ndmp9.h"

bool_t
xdr_ndmp9_error (XDR *xdrs, ndmp9_error *objp)
{
	register int32_t *buf;

	 if (!xdr_enum (xdrs, (enum_t *) objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_message (XDR *xdrs, ndmp9_message *objp)
{
	register int32_t *buf;

	 if (!xdr_enum (xdrs, (enum_t *) objp))
		 return FALSE;
	return TRUE;
}
extern bool_t xdr_ndmp9_no_arguments();
#define ndmp9_no_arguments int

bool_t
xdr_ndmp9_just_error_reply (XDR *xdrs, ndmp9_just_error_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_error (xdrs, &objp->error))
		 return FALSE;
	return TRUE;
}
extern bool_t xdr_ndmp9_u_quad();
#define ndmp9_u_quad unsigned long long

bool_t
xdr_ndmp9_validity (XDR *xdrs, ndmp9_validity *objp)
{
	register int32_t *buf;

	 if (!xdr_enum (xdrs, (enum_t *) objp))
		 return FALSE;
	return TRUE;
}
#define NDMP9_INVALID_U_LONG 0xFFFFFFFFul

bool_t
xdr_ndmp9_valid_u_long (XDR *xdrs, ndmp9_valid_u_long *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_validity (xdrs, &objp->valid))
		 return FALSE;
	 if (!xdr_u_long (xdrs, &objp->value))
		 return FALSE;
	return TRUE;
}
#define NDMP9_INVALID_U_QUAD 0xFFFFFFFFFFFFFFFFull

bool_t
xdr_ndmp9_valid_u_quad (XDR *xdrs, ndmp9_valid_u_quad *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_validity (xdrs, &objp->valid))
		 return FALSE;
	 if (!xdr_ndmp9_u_quad (xdrs, &objp->value))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_pval (XDR *xdrs, ndmp9_pval *objp)
{
	register int32_t *buf;

	 if (!xdr_string (xdrs, &objp->name, ~0))
		 return FALSE;
	 if (!xdr_string (xdrs, &objp->value, ~0))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_auth_type (XDR *xdrs, ndmp9_auth_type *objp)
{
	register int32_t *buf;

	 if (!xdr_enum (xdrs, (enum_t *) objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_auth_attr (XDR *xdrs, ndmp9_auth_attr *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_auth_type (xdrs, &objp->auth_type))
		 return FALSE;
	switch (objp->auth_type) {
	case NDMP9_AUTH_NONE:
		break;
	case NDMP9_AUTH_TEXT:
		break;
	case NDMP9_AUTH_MD5:
		 if (!xdr_opaque (xdrs, objp->ndmp9_auth_attr_u.challenge, 64))
			 return FALSE;
		break;
	default:
		return FALSE;
	}
	return TRUE;
}

bool_t
xdr_ndmp9_auth_text (XDR *xdrs, ndmp9_auth_text *objp)
{
	register int32_t *buf;

	 if (!xdr_string (xdrs, &objp->auth_id, ~0))
		 return FALSE;
	 if (!xdr_string (xdrs, &objp->auth_password, ~0))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_auth_md5 (XDR *xdrs, ndmp9_auth_md5 *objp)
{
	register int32_t *buf;

	int i;
	 if (!xdr_string (xdrs, &objp->auth_id, ~0))
		 return FALSE;
	 if (!xdr_opaque (xdrs, objp->auth_digest, 16))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_auth_data (XDR *xdrs, ndmp9_auth_data *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_auth_type (xdrs, &objp->auth_type))
		 return FALSE;
	switch (objp->auth_type) {
	case NDMP9_AUTH_NONE:
		break;
	case NDMP9_AUTH_TEXT:
		 if (!xdr_ndmp9_auth_text (xdrs, &objp->ndmp9_auth_data_u.auth_text))
			 return FALSE;
		break;
	case NDMP9_AUTH_MD5:
		 if (!xdr_ndmp9_auth_md5 (xdrs, &objp->ndmp9_auth_data_u.auth_md5))
			 return FALSE;
		break;
	default:
		return FALSE;
	}
	return TRUE;
}

bool_t
xdr_ndmp9_addr_type (XDR *xdrs, ndmp9_addr_type *objp)
{
	register int32_t *buf;

	 if (!xdr_enum (xdrs, (enum_t *) objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_tcp_addr (XDR *xdrs, ndmp9_tcp_addr *objp)
{
	register int32_t *buf;

	 if (!xdr_u_long (xdrs, &objp->ip_addr))
		 return FALSE;
	 if (!xdr_u_short (xdrs, &objp->port))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_addr (XDR *xdrs, ndmp9_addr *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_addr_type (xdrs, &objp->addr_type))
		 return FALSE;
	switch (objp->addr_type) {
	case NDMP9_ADDR_LOCAL:
	case NDMP9_ADDR_AS_CONNECTED:
		break;
	case NDMP9_ADDR_TCP:
		 if (!xdr_ndmp9_tcp_addr (xdrs, &objp->ndmp9_addr_u.tcp_addr))
			 return FALSE;
		break;
	default:
		return FALSE;
	}
	return TRUE;
}

bool_t
xdr_ndmp9_connect_open_request (XDR *xdrs, ndmp9_connect_open_request *objp)
{
	register int32_t *buf;

	 if (!xdr_u_short (xdrs, &objp->protocol_version))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_connect_open_reply (XDR *xdrs, ndmp9_connect_open_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_just_error_reply (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_connect_client_auth_request (XDR *xdrs, ndmp9_connect_client_auth_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_auth_data (xdrs, &objp->auth_data))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_connect_client_auth_reply (XDR *xdrs, ndmp9_connect_client_auth_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_just_error_reply (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_connect_close_request (XDR *xdrs, ndmp9_connect_close_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_no_arguments (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_connect_close_reply (XDR *xdrs, ndmp9_connect_close_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_no_arguments (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_connect_server_auth_request (XDR *xdrs, ndmp9_connect_server_auth_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_auth_attr (xdrs, &objp->client_attr))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_connect_server_auth_reply (XDR *xdrs, ndmp9_connect_server_auth_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_error (xdrs, &objp->error))
		 return FALSE;
	 if (!xdr_ndmp9_auth_data (xdrs, &objp->server_result))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_butype_info (XDR *xdrs, ndmp9_butype_info *objp)
{
	register int32_t *buf;

	 if (!xdr_string (xdrs, &objp->butype_name, ~0))
		 return FALSE;
	 if (!xdr_ndmp9_valid_u_long (xdrs, &objp->v2attr))
		 return FALSE;
	 if (!xdr_ndmp9_valid_u_long (xdrs, &objp->v3attr))
		 return FALSE;
	 if (!xdr_ndmp9_valid_u_long (xdrs, &objp->v4attr))
		 return FALSE;
	 if (!xdr_array (xdrs, (char **)&objp->default_env.default_env_val, (u_int *) &objp->default_env.default_env_len, ~0,
		sizeof (ndmp9_pval), (xdrproc_t) xdr_ndmp9_pval))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_fs_info (XDR *xdrs, ndmp9_fs_info *objp)
{
	register int32_t *buf;

	 if (!xdr_string (xdrs, &objp->fs_type, ~0))
		 return FALSE;
	 if (!xdr_string (xdrs, &objp->fs_logical_device, ~0))
		 return FALSE;
	 if (!xdr_string (xdrs, &objp->fs_physical_device, ~0))
		 return FALSE;
	 if (!xdr_ndmp9_valid_u_quad (xdrs, &objp->total_size))
		 return FALSE;
	 if (!xdr_ndmp9_valid_u_quad (xdrs, &objp->used_size))
		 return FALSE;
	 if (!xdr_ndmp9_valid_u_quad (xdrs, &objp->avail_size))
		 return FALSE;
	 if (!xdr_ndmp9_valid_u_quad (xdrs, &objp->total_inodes))
		 return FALSE;
	 if (!xdr_ndmp9_valid_u_quad (xdrs, &objp->used_inodes))
		 return FALSE;
	 if (!xdr_array (xdrs, (char **)&objp->fs_env.fs_env_val, (u_int *) &objp->fs_env.fs_env_len, ~0,
		sizeof (ndmp9_pval), (xdrproc_t) xdr_ndmp9_pval))
		 return FALSE;
	 if (!xdr_string (xdrs, &objp->fs_status, ~0))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_device_capability (XDR *xdrs, ndmp9_device_capability *objp)
{
	register int32_t *buf;

	 if (!xdr_string (xdrs, &objp->device, ~0))
		 return FALSE;
	 if (!xdr_ndmp9_valid_u_long (xdrs, &objp->v3attr))
		 return FALSE;
	 if (!xdr_ndmp9_valid_u_long (xdrs, &objp->v4attr))
		 return FALSE;
	 if (!xdr_array (xdrs, (char **)&objp->capability.capability_val, (u_int *) &objp->capability.capability_len, ~0,
		sizeof (ndmp9_pval), (xdrproc_t) xdr_ndmp9_pval))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_device_info (XDR *xdrs, ndmp9_device_info *objp)
{
	register int32_t *buf;

	 if (!xdr_string (xdrs, &objp->model, ~0))
		 return FALSE;
	 if (!xdr_array (xdrs, (char **)&objp->caplist.caplist_val, (u_int *) &objp->caplist.caplist_len, ~0,
		sizeof (ndmp9_device_capability), (xdrproc_t) xdr_ndmp9_device_capability))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_config_info (XDR *xdrs, ndmp9_config_info *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_error (xdrs, &objp->error))
		 return FALSE;
	 if (!xdr_string (xdrs, &objp->hostname, ~0))
		 return FALSE;
	 if (!xdr_string (xdrs, &objp->os_type, ~0))
		 return FALSE;
	 if (!xdr_string (xdrs, &objp->os_vers, ~0))
		 return FALSE;
	 if (!xdr_string (xdrs, &objp->hostid, ~0))
		 return FALSE;
	 if (!xdr_string (xdrs, &objp->vendor_name, ~0))
		 return FALSE;
	 if (!xdr_string (xdrs, &objp->product_name, ~0))
		 return FALSE;
	 if (!xdr_string (xdrs, &objp->revision_number, ~0))
		 return FALSE;
	 if (!xdr_u_long (xdrs, &objp->authtypes))
		 return FALSE;
	 if (!xdr_u_long (xdrs, &objp->conntypes))
		 return FALSE;
	 if (!xdr_array (xdrs, (char **)&objp->butype_info.butype_info_val, (u_int *) &objp->butype_info.butype_info_len, ~0,
		sizeof (ndmp9_butype_info), (xdrproc_t) xdr_ndmp9_butype_info))
		 return FALSE;
	 if (!xdr_array (xdrs, (char **)&objp->fs_info.fs_info_val, (u_int *) &objp->fs_info.fs_info_len, ~0,
		sizeof (ndmp9_fs_info), (xdrproc_t) xdr_ndmp9_fs_info))
		 return FALSE;
	 if (!xdr_array (xdrs, (char **)&objp->tape_info.tape_info_val, (u_int *) &objp->tape_info.tape_info_len, ~0,
		sizeof (ndmp9_device_info), (xdrproc_t) xdr_ndmp9_device_info))
		 return FALSE;
	 if (!xdr_array (xdrs, (char **)&objp->scsi_info.scsi_info_val, (u_int *) &objp->scsi_info.scsi_info_len, ~0,
		sizeof (ndmp9_device_info), (xdrproc_t) xdr_ndmp9_device_info))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_config_get_info_request (XDR *xdrs, ndmp9_config_get_info_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_no_arguments (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_config_get_info_reply (XDR *xdrs, ndmp9_config_get_info_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_error (xdrs, &objp->error))
		 return FALSE;
	 if (!xdr_ndmp9_config_info (xdrs, &objp->config_info))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_config_get_host_info_request (XDR *xdrs, ndmp9_config_get_host_info_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_no_arguments (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_config_get_host_info_reply (XDR *xdrs, ndmp9_config_get_host_info_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_config_get_info_reply (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_config_get_connection_type_request (XDR *xdrs, ndmp9_config_get_connection_type_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_no_arguments (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_config_get_connection_type_reply (XDR *xdrs, ndmp9_config_get_connection_type_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_config_get_info_reply (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_config_get_server_info_request (XDR *xdrs, ndmp9_config_get_server_info_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_no_arguments (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_config_get_server_info_reply (XDR *xdrs, ndmp9_config_get_server_info_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_config_get_info_reply (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_config_get_butype_info_request (XDR *xdrs, ndmp9_config_get_butype_info_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_no_arguments (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_config_get_butype_info_reply (XDR *xdrs, ndmp9_config_get_butype_info_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_config_get_info_reply (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_config_get_fs_info_request (XDR *xdrs, ndmp9_config_get_fs_info_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_no_arguments (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_config_get_fs_info_reply (XDR *xdrs, ndmp9_config_get_fs_info_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_config_get_info_reply (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_config_get_tape_info_request (XDR *xdrs, ndmp9_config_get_tape_info_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_no_arguments (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_config_get_tape_info_reply (XDR *xdrs, ndmp9_config_get_tape_info_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_config_get_info_reply (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_config_get_scsi_info_request (XDR *xdrs, ndmp9_config_get_scsi_info_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_no_arguments (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_config_get_scsi_info_reply (XDR *xdrs, ndmp9_config_get_scsi_info_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_config_get_info_reply (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_config_get_auth_attr_request (XDR *xdrs, ndmp9_config_get_auth_attr_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_auth_type (xdrs, &objp->auth_type))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_config_get_auth_attr_reply (XDR *xdrs, ndmp9_config_get_auth_attr_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_error (xdrs, &objp->error))
		 return FALSE;
	 if (!xdr_ndmp9_auth_attr (xdrs, &objp->server_attr))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_scsi_open_request (XDR *xdrs, ndmp9_scsi_open_request *objp)
{
	register int32_t *buf;

	 if (!xdr_string (xdrs, &objp->device, ~0))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_scsi_open_reply (XDR *xdrs, ndmp9_scsi_open_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_just_error_reply (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_scsi_close_request (XDR *xdrs, ndmp9_scsi_close_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_no_arguments (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_scsi_close_reply (XDR *xdrs, ndmp9_scsi_close_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_just_error_reply (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_scsi_get_state_request (XDR *xdrs, ndmp9_scsi_get_state_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_no_arguments (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_scsi_get_state_reply (XDR *xdrs, ndmp9_scsi_get_state_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_error (xdrs, &objp->error))
		 return FALSE;
	 if (!xdr_short (xdrs, &objp->target_controller))
		 return FALSE;
	 if (!xdr_short (xdrs, &objp->target_id))
		 return FALSE;
	 if (!xdr_short (xdrs, &objp->target_lun))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_scsi_set_target_request (XDR *xdrs, ndmp9_scsi_set_target_request *objp)
{
	register int32_t *buf;

	 if (!xdr_string (xdrs, &objp->device, ~0))
		 return FALSE;
	 if (!xdr_u_short (xdrs, &objp->target_controller))
		 return FALSE;
	 if (!xdr_u_short (xdrs, &objp->target_id))
		 return FALSE;
	 if (!xdr_u_short (xdrs, &objp->target_lun))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_scsi_set_target_reply (XDR *xdrs, ndmp9_scsi_set_target_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_just_error_reply (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_scsi_reset_device_request (XDR *xdrs, ndmp9_scsi_reset_device_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_no_arguments (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_scsi_reset_device_reply (XDR *xdrs, ndmp9_scsi_reset_device_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_just_error_reply (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_scsi_reset_bus_request (XDR *xdrs, ndmp9_scsi_reset_bus_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_no_arguments (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_scsi_reset_bus_reply (XDR *xdrs, ndmp9_scsi_reset_bus_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_just_error_reply (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_scsi_data_dir (XDR *xdrs, ndmp9_scsi_data_dir *objp)
{
	register int32_t *buf;

	 if (!xdr_enum (xdrs, (enum_t *) objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_execute_cdb_request (XDR *xdrs, ndmp9_execute_cdb_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_scsi_data_dir (xdrs, &objp->data_dir))
		 return FALSE;
	 if (!xdr_u_long (xdrs, &objp->timeout))
		 return FALSE;
	 if (!xdr_u_long (xdrs, &objp->datain_len))
		 return FALSE;
	 if (!xdr_bytes (xdrs, (char **)&objp->cdb.cdb_val, (u_int *) &objp->cdb.cdb_len, ~0))
		 return FALSE;
	 if (!xdr_bytes (xdrs, (char **)&objp->dataout.dataout_val, (u_int *) &objp->dataout.dataout_len, ~0))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_execute_cdb_reply (XDR *xdrs, ndmp9_execute_cdb_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_error (xdrs, &objp->error))
		 return FALSE;
	 if (!xdr_u_char (xdrs, &objp->status))
		 return FALSE;
	 if (!xdr_u_long (xdrs, &objp->dataout_len))
		 return FALSE;
	 if (!xdr_bytes (xdrs, (char **)&objp->datain.datain_val, (u_int *) &objp->datain.datain_len, ~0))
		 return FALSE;
	 if (!xdr_bytes (xdrs, (char **)&objp->ext_sense.ext_sense_val, (u_int *) &objp->ext_sense.ext_sense_len, ~0))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_scsi_execute_cdb_request (XDR *xdrs, ndmp9_scsi_execute_cdb_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_execute_cdb_request (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_scsi_execute_cdb_reply (XDR *xdrs, ndmp9_scsi_execute_cdb_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_execute_cdb_reply (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_tape_open_mode (XDR *xdrs, ndmp9_tape_open_mode *objp)
{
	register int32_t *buf;

	 if (!xdr_enum (xdrs, (enum_t *) objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_tape_state (XDR *xdrs, ndmp9_tape_state *objp)
{
	register int32_t *buf;

	 if (!xdr_enum (xdrs, (enum_t *) objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_tape_open_request (XDR *xdrs, ndmp9_tape_open_request *objp)
{
	register int32_t *buf;

	 if (!xdr_string (xdrs, &objp->device, ~0))
		 return FALSE;
	 if (!xdr_ndmp9_tape_open_mode (xdrs, &objp->mode))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_tape_open_reply (XDR *xdrs, ndmp9_tape_open_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_just_error_reply (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_tape_close_request (XDR *xdrs, ndmp9_tape_close_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_no_arguments (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_tape_close_reply (XDR *xdrs, ndmp9_tape_close_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_just_error_reply (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_tape_get_state_request (XDR *xdrs, ndmp9_tape_get_state_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_no_arguments (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_tape_get_state_reply (XDR *xdrs, ndmp9_tape_get_state_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_error (xdrs, &objp->error))
		 return FALSE;
	 if (!xdr_u_long (xdrs, &objp->flags))
		 return FALSE;
	 if (!xdr_ndmp9_tape_state (xdrs, &objp->state))
		 return FALSE;
	 if (!xdr_ndmp9_tape_open_mode (xdrs, &objp->open_mode))
		 return FALSE;
	 if (!xdr_ndmp9_valid_u_long (xdrs, &objp->file_num))
		 return FALSE;
	 if (!xdr_ndmp9_valid_u_long (xdrs, &objp->soft_errors))
		 return FALSE;
	 if (!xdr_ndmp9_valid_u_long (xdrs, &objp->block_size))
		 return FALSE;
	 if (!xdr_ndmp9_valid_u_long (xdrs, &objp->blockno))
		 return FALSE;
	 if (!xdr_ndmp9_valid_u_quad (xdrs, &objp->total_space))
		 return FALSE;
	 if (!xdr_ndmp9_valid_u_quad (xdrs, &objp->space_remain))
		 return FALSE;
	 if (!xdr_ndmp9_valid_u_long (xdrs, &objp->partition))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_tape_mtio_op (XDR *xdrs, ndmp9_tape_mtio_op *objp)
{
	register int32_t *buf;

	 if (!xdr_enum (xdrs, (enum_t *) objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_tape_mtio_request (XDR *xdrs, ndmp9_tape_mtio_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_tape_mtio_op (xdrs, &objp->tape_op))
		 return FALSE;
	 if (!xdr_u_long (xdrs, &objp->count))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_tape_mtio_reply (XDR *xdrs, ndmp9_tape_mtio_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_error (xdrs, &objp->error))
		 return FALSE;
	 if (!xdr_u_long (xdrs, &objp->resid_count))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_tape_write_request (XDR *xdrs, ndmp9_tape_write_request *objp)
{
	register int32_t *buf;

	 if (!xdr_bytes (xdrs, (char **)&objp->data_out.data_out_val, (u_int *) &objp->data_out.data_out_len, ~0))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_tape_write_reply (XDR *xdrs, ndmp9_tape_write_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_error (xdrs, &objp->error))
		 return FALSE;
	 if (!xdr_u_long (xdrs, &objp->count))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_tape_read_request (XDR *xdrs, ndmp9_tape_read_request *objp)
{
	register int32_t *buf;

	 if (!xdr_u_long (xdrs, &objp->count))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_tape_read_reply (XDR *xdrs, ndmp9_tape_read_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_error (xdrs, &objp->error))
		 return FALSE;
	 if (!xdr_bytes (xdrs, (char **)&objp->data_in.data_in_val, (u_int *) &objp->data_in.data_in_len, ~0))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_tape_execute_cdb_request (XDR *xdrs, ndmp9_tape_execute_cdb_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_execute_cdb_request (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_tape_execute_cdb_reply (XDR *xdrs, ndmp9_tape_execute_cdb_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_execute_cdb_reply (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_mover_state (XDR *xdrs, ndmp9_mover_state *objp)
{
	register int32_t *buf;

	 if (!xdr_enum (xdrs, (enum_t *) objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_mover_mode (XDR *xdrs, ndmp9_mover_mode *objp)
{
	register int32_t *buf;

	 if (!xdr_enum (xdrs, (enum_t *) objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_mover_pause_reason (XDR *xdrs, ndmp9_mover_pause_reason *objp)
{
	register int32_t *buf;

	 if (!xdr_enum (xdrs, (enum_t *) objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_mover_halt_reason (XDR *xdrs, ndmp9_mover_halt_reason *objp)
{
	register int32_t *buf;

	 if (!xdr_enum (xdrs, (enum_t *) objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_mover_get_state_request (XDR *xdrs, ndmp9_mover_get_state_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_no_arguments (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_mover_get_state_reply (XDR *xdrs, ndmp9_mover_get_state_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_error (xdrs, &objp->error))
		 return FALSE;
	 if (!xdr_ndmp9_mover_state (xdrs, &objp->state))
		 return FALSE;
	 if (!xdr_ndmp9_mover_mode (xdrs, &objp->mode))
		 return FALSE;
	 if (!xdr_ndmp9_mover_pause_reason (xdrs, &objp->pause_reason))
		 return FALSE;
	 if (!xdr_ndmp9_mover_halt_reason (xdrs, &objp->halt_reason))
		 return FALSE;
	 if (!xdr_u_long (xdrs, &objp->record_size))
		 return FALSE;
	 if (!xdr_u_long (xdrs, &objp->record_num))
		 return FALSE;
	 if (!xdr_ndmp9_u_quad (xdrs, &objp->bytes_moved))
		 return FALSE;
	 if (!xdr_ndmp9_u_quad (xdrs, &objp->seek_position))
		 return FALSE;
	 if (!xdr_ndmp9_u_quad (xdrs, &objp->bytes_left_to_read))
		 return FALSE;
	 if (!xdr_ndmp9_u_quad (xdrs, &objp->window_offset))
		 return FALSE;
	 if (!xdr_ndmp9_u_quad (xdrs, &objp->window_length))
		 return FALSE;
	 if (!xdr_ndmp9_addr (xdrs, &objp->data_connection_addr))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_mover_listen_request (XDR *xdrs, ndmp9_mover_listen_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_mover_mode (xdrs, &objp->mode))
		 return FALSE;
	 if (!xdr_ndmp9_addr_type (xdrs, &objp->addr_type))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_mover_listen_reply (XDR *xdrs, ndmp9_mover_listen_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_error (xdrs, &objp->error))
		 return FALSE;
	 if (!xdr_ndmp9_addr (xdrs, &objp->data_connection_addr))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_mover_connect_request (XDR *xdrs, ndmp9_mover_connect_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_mover_mode (xdrs, &objp->mode))
		 return FALSE;
	 if (!xdr_ndmp9_addr (xdrs, &objp->addr))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_mover_connect_reply (XDR *xdrs, ndmp9_mover_connect_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_just_error_reply (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_mover_set_record_size_request (XDR *xdrs, ndmp9_mover_set_record_size_request *objp)
{
	register int32_t *buf;

	 if (!xdr_u_long (xdrs, &objp->record_size))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_mover_set_record_size_reply (XDR *xdrs, ndmp9_mover_set_record_size_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_just_error_reply (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_mover_set_window_request (XDR *xdrs, ndmp9_mover_set_window_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_u_quad (xdrs, &objp->offset))
		 return FALSE;
	 if (!xdr_ndmp9_u_quad (xdrs, &objp->length))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_mover_set_window_reply (XDR *xdrs, ndmp9_mover_set_window_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_just_error_reply (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_mover_continue_request (XDR *xdrs, ndmp9_mover_continue_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_no_arguments (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_mover_continue_reply (XDR *xdrs, ndmp9_mover_continue_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_just_error_reply (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_mover_abort_request (XDR *xdrs, ndmp9_mover_abort_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_no_arguments (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_mover_abort_reply (XDR *xdrs, ndmp9_mover_abort_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_just_error_reply (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_mover_stop_request (XDR *xdrs, ndmp9_mover_stop_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_no_arguments (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_mover_stop_reply (XDR *xdrs, ndmp9_mover_stop_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_just_error_reply (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_mover_read_request (XDR *xdrs, ndmp9_mover_read_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_u_quad (xdrs, &objp->offset))
		 return FALSE;
	 if (!xdr_ndmp9_u_quad (xdrs, &objp->length))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_mover_read_reply (XDR *xdrs, ndmp9_mover_read_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_just_error_reply (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_mover_close_request (XDR *xdrs, ndmp9_mover_close_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_no_arguments (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_mover_close_reply (XDR *xdrs, ndmp9_mover_close_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_just_error_reply (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_data_operation (XDR *xdrs, ndmp9_data_operation *objp)
{
	register int32_t *buf;

	 if (!xdr_enum (xdrs, (enum_t *) objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_data_state (XDR *xdrs, ndmp9_data_state *objp)
{
	register int32_t *buf;

	 if (!xdr_enum (xdrs, (enum_t *) objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_data_halt_reason (XDR *xdrs, ndmp9_data_halt_reason *objp)
{
	register int32_t *buf;

	 if (!xdr_enum (xdrs, (enum_t *) objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_data_get_state_request (XDR *xdrs, ndmp9_data_get_state_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_no_arguments (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_data_get_state_reply (XDR *xdrs, ndmp9_data_get_state_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_error (xdrs, &objp->error))
		 return FALSE;
	 if (!xdr_ndmp9_data_operation (xdrs, &objp->operation))
		 return FALSE;
	 if (!xdr_ndmp9_data_state (xdrs, &objp->state))
		 return FALSE;
	 if (!xdr_ndmp9_data_halt_reason (xdrs, &objp->halt_reason))
		 return FALSE;
	 if (!xdr_ndmp9_u_quad (xdrs, &objp->bytes_processed))
		 return FALSE;
	 if (!xdr_ndmp9_valid_u_quad (xdrs, &objp->est_bytes_remain))
		 return FALSE;
	 if (!xdr_ndmp9_valid_u_long (xdrs, &objp->est_time_remain))
		 return FALSE;
	 if (!xdr_ndmp9_addr (xdrs, &objp->data_connection_addr))
		 return FALSE;
	 if (!xdr_ndmp9_u_quad (xdrs, &objp->read_offset))
		 return FALSE;
	 if (!xdr_ndmp9_u_quad (xdrs, &objp->read_length))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_name (XDR *xdrs, ndmp9_name *objp)
{
	register int32_t *buf;

	 if (!xdr_string (xdrs, &objp->original_path, ~0))
		 return FALSE;
	 if (!xdr_string (xdrs, &objp->destination_path, ~0))
		 return FALSE;
	 if (!xdr_ndmp9_valid_u_quad (xdrs, &objp->fh_info))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_data_start_backup_request (XDR *xdrs, ndmp9_data_start_backup_request *objp)
{
	register int32_t *buf;

	 if (!xdr_string (xdrs, &objp->bu_type, ~0))
		 return FALSE;
	 if (!xdr_array (xdrs, (char **)&objp->env.env_val, (u_int *) &objp->env.env_len, ~0,
		sizeof (ndmp9_pval), (xdrproc_t) xdr_ndmp9_pval))
		 return FALSE;
	 if (!xdr_ndmp9_addr (xdrs, &objp->addr))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_data_start_backup_reply (XDR *xdrs, ndmp9_data_start_backup_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_just_error_reply (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_data_start_recover_request (XDR *xdrs, ndmp9_data_start_recover_request *objp)
{
	register int32_t *buf;

	 if (!xdr_array (xdrs, (char **)&objp->env.env_val, (u_int *) &objp->env.env_len, ~0,
		sizeof (ndmp9_pval), (xdrproc_t) xdr_ndmp9_pval))
		 return FALSE;
	 if (!xdr_array (xdrs, (char **)&objp->nlist.nlist_val, (u_int *) &objp->nlist.nlist_len, ~0,
		sizeof (ndmp9_name), (xdrproc_t) xdr_ndmp9_name))
		 return FALSE;
	 if (!xdr_string (xdrs, &objp->bu_type, ~0))
		 return FALSE;
	 if (!xdr_ndmp9_addr (xdrs, &objp->addr))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_data_start_recover_reply (XDR *xdrs, ndmp9_data_start_recover_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_just_error_reply (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_data_start_recover_filehist_request (XDR *xdrs, ndmp9_data_start_recover_filehist_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_data_start_recover_request (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_data_start_recover_filehist_reply (XDR *xdrs, ndmp9_data_start_recover_filehist_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_data_start_recover_reply (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_data_abort_request (XDR *xdrs, ndmp9_data_abort_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_no_arguments (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_data_abort_reply (XDR *xdrs, ndmp9_data_abort_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_just_error_reply (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_data_stop_request (XDR *xdrs, ndmp9_data_stop_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_no_arguments (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_data_stop_reply (XDR *xdrs, ndmp9_data_stop_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_just_error_reply (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_data_get_env_request (XDR *xdrs, ndmp9_data_get_env_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_no_arguments (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_data_get_env_reply (XDR *xdrs, ndmp9_data_get_env_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_error (xdrs, &objp->error))
		 return FALSE;
	 if (!xdr_array (xdrs, (char **)&objp->env.env_val, (u_int *) &objp->env.env_len, ~0,
		sizeof (ndmp9_pval), (xdrproc_t) xdr_ndmp9_pval))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_data_listen_request (XDR *xdrs, ndmp9_data_listen_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_addr_type (xdrs, &objp->addr_type))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_data_listen_reply (XDR *xdrs, ndmp9_data_listen_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_error (xdrs, &objp->error))
		 return FALSE;
	 if (!xdr_ndmp9_addr (xdrs, &objp->data_connection_addr))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_data_connect_request (XDR *xdrs, ndmp9_data_connect_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_addr (xdrs, &objp->addr))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_data_connect_reply (XDR *xdrs, ndmp9_data_connect_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_just_error_reply (xdrs, objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_notify_data_halted_request (XDR *xdrs, ndmp9_notify_data_halted_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_data_halt_reason (xdrs, &objp->reason))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_connect_reason (XDR *xdrs, ndmp9_connect_reason *objp)
{
	register int32_t *buf;

	 if (!xdr_enum (xdrs, (enum_t *) objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_notify_connected_request (XDR *xdrs, ndmp9_notify_connected_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_connect_reason (xdrs, &objp->reason))
		 return FALSE;
	 if (!xdr_u_short (xdrs, &objp->protocol_version))
		 return FALSE;
	 if (!xdr_string (xdrs, &objp->text_reason, ~0))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_notify_mover_halted_request (XDR *xdrs, ndmp9_notify_mover_halted_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_mover_halt_reason (xdrs, &objp->reason))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_notify_mover_paused_request (XDR *xdrs, ndmp9_notify_mover_paused_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_mover_pause_reason (xdrs, &objp->reason))
		 return FALSE;
	 if (!xdr_ndmp9_u_quad (xdrs, &objp->seek_position))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_notify_data_read_request (XDR *xdrs, ndmp9_notify_data_read_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_u_quad (xdrs, &objp->offset))
		 return FALSE;
	 if (!xdr_ndmp9_u_quad (xdrs, &objp->length))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_log_type (XDR *xdrs, ndmp9_log_type *objp)
{
	register int32_t *buf;

	 if (!xdr_enum (xdrs, (enum_t *) objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_log_message_request (XDR *xdrs, ndmp9_log_message_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_log_type (xdrs, &objp->log_type))
		 return FALSE;
	 if (!xdr_u_long (xdrs, &objp->message_id))
		 return FALSE;
	 if (!xdr_string (xdrs, &objp->entry, ~0))
		 return FALSE;
	 if (!xdr_ndmp9_valid_u_long (xdrs, &objp->associated_message_sequence))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_recovery_status (XDR *xdrs, ndmp9_recovery_status *objp)
{
	register int32_t *buf;

	 if (!xdr_enum (xdrs, (enum_t *) objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_log_file_request (XDR *xdrs, ndmp9_log_file_request *objp)
{
	register int32_t *buf;

	 if (!xdr_string (xdrs, &objp->name, ~0))
		 return FALSE;
	 if (!xdr_ndmp9_recovery_status (xdrs, &objp->recovery_status))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_file_type (XDR *xdrs, ndmp9_file_type *objp)
{
	register int32_t *buf;

	 if (!xdr_enum (xdrs, (enum_t *) objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_file_stat (XDR *xdrs, ndmp9_file_stat *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_file_type (xdrs, &objp->ftype))
		 return FALSE;
	 if (!xdr_ndmp9_valid_u_long (xdrs, &objp->mtime))
		 return FALSE;
	 if (!xdr_ndmp9_valid_u_long (xdrs, &objp->atime))
		 return FALSE;
	 if (!xdr_ndmp9_valid_u_long (xdrs, &objp->ctime))
		 return FALSE;
	 if (!xdr_ndmp9_valid_u_long (xdrs, &objp->uid))
		 return FALSE;
	 if (!xdr_ndmp9_valid_u_long (xdrs, &objp->gid))
		 return FALSE;
	 if (!xdr_ndmp9_valid_u_long (xdrs, &objp->mode))
		 return FALSE;
	 if (!xdr_ndmp9_valid_u_quad (xdrs, &objp->size))
		 return FALSE;
	 if (!xdr_ndmp9_valid_u_long (xdrs, &objp->links))
		 return FALSE;
	 if (!xdr_ndmp9_valid_u_quad (xdrs, &objp->node))
		 return FALSE;
	 if (!xdr_ndmp9_valid_u_quad (xdrs, &objp->fh_info))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_file (XDR *xdrs, ndmp9_file *objp)
{
	register int32_t *buf;

	 if (!xdr_string (xdrs, &objp->unix_path, ~0))
		 return FALSE;
	 if (!xdr_ndmp9_file_stat (xdrs, &objp->fstat))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_fh_add_file_request (XDR *xdrs, ndmp9_fh_add_file_request *objp)
{
	register int32_t *buf;

	 if (!xdr_array (xdrs, (char **)&objp->files.files_val, (u_int *) &objp->files.files_len, ~0,
		sizeof (ndmp9_file), (xdrproc_t) xdr_ndmp9_file))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_dir (XDR *xdrs, ndmp9_dir *objp)
{
	register int32_t *buf;

	 if (!xdr_string (xdrs, &objp->unix_name, ~0))
		 return FALSE;
	 if (!xdr_ndmp9_u_quad (xdrs, &objp->node))
		 return FALSE;
	 if (!xdr_ndmp9_u_quad (xdrs, &objp->parent))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_fh_add_dir_request (XDR *xdrs, ndmp9_fh_add_dir_request *objp)
{
	register int32_t *buf;

	 if (!xdr_array (xdrs, (char **)&objp->dirs.dirs_val, (u_int *) &objp->dirs.dirs_len, ~0,
		sizeof (ndmp9_dir), (xdrproc_t) xdr_ndmp9_dir))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_node (XDR *xdrs, ndmp9_node *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp9_file_stat (xdrs, &objp->fstat))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp9_fh_add_node_request (XDR *xdrs, ndmp9_fh_add_node_request *objp)
{
	register int32_t *buf;

	 if (!xdr_array (xdrs, (char **)&objp->nodes.nodes_val, (u_int *) &objp->nodes.nodes_len, ~0,
		sizeof (ndmp9_node), (xdrproc_t) xdr_ndmp9_node))
		 return FALSE;
	return TRUE;
}
