/*
 * Please do not edit this file.
 * It was generated using rpcgen.
 */

#include "config.h"
#include "ndmp0.h"

bool_t
xdr_ndmp0_error (XDR *xdrs, ndmp0_error *objp)
{
	register int32_t *buf;

	 if (!xdr_enum (xdrs, (enum_t *) objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp0_header_message_type (XDR *xdrs, ndmp0_header_message_type *objp)
{
	register int32_t *buf;

	 if (!xdr_enum (xdrs, (enum_t *) objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp0_message (XDR *xdrs, ndmp0_message *objp)
{
	register int32_t *buf;

	 if (!xdr_enum (xdrs, (enum_t *) objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp0_header (XDR *xdrs, ndmp0_header *objp)
{
	register int32_t *buf;

	 if (!xdr_u_long (xdrs, &objp->sequence))
		 return FALSE;
	 if (!xdr_u_long (xdrs, &objp->time_stamp))
		 return FALSE;
	 if (!xdr_ndmp0_header_message_type (xdrs, &objp->message_type))
		 return FALSE;
	 if (!xdr_ndmp0_message (xdrs, &objp->message))
		 return FALSE;
	 if (!xdr_u_long (xdrs, &objp->reply_sequence))
		 return FALSE;
	 if (!xdr_ndmp0_error (xdrs, &objp->error))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp0_connect_open_request (XDR *xdrs, ndmp0_connect_open_request *objp)
{
	register int32_t *buf;

	 if (!xdr_u_short (xdrs, &objp->protocol_version))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp0_connect_open_reply (XDR *xdrs, ndmp0_connect_open_reply *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp0_error (xdrs, &objp->error))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp0_connect_reason (XDR *xdrs, ndmp0_connect_reason *objp)
{
	register int32_t *buf;

	 if (!xdr_enum (xdrs, (enum_t *) objp))
		 return FALSE;
	return TRUE;
}

bool_t
xdr_ndmp0_notify_connected_request (XDR *xdrs, ndmp0_notify_connected_request *objp)
{
	register int32_t *buf;

	 if (!xdr_ndmp0_connect_reason (xdrs, &objp->reason))
		 return FALSE;
	 if (!xdr_u_short (xdrs, &objp->protocol_version))
		 return FALSE;
	 if (!xdr_string (xdrs, &objp->text_reason, ~0))
		 return FALSE;
	return TRUE;
}
